from enum import Enum
from pathlib import Path

import typer


class AccessPolicy(str, Enum):
    UNRESTRICTED = "NOT_SECURE"
    CERTIFICATE = "CERTIFICATE"
    TOKEN = "TOKEN"


CoreSystemAddress = typer.Option("127.0.0.1", "--address", "-a")


def CoreSystemPort(port: int):
    return typer.Option(port, "--port", "-p")


CertDirectory = typer.Option(Path("."), "--cert-dir")
ARG_SERVICE_DEFINITION = typer.Argument(..., metavar="SERVICE_DEFINITION", help="")
ARG_SERVICE_URI = typer.Argument(..., metavar="SERVICE_URI", help="")
ARG_SERVICE_INTERFACE = typer.Argument(
    ...,
    metavar="SERVICE_INTERFACE",
    help="Must be of format <PROTOCOL>-<SECURITY>-<PAYLOAD>.",
)


def system_option(
    system_name: str,
):
    return typer.Option(
        (None, None, None),
        show_default=False,
        metavar="SYSTEM_NAME ADDRESS PORT",
        help=f"{system_name.capitalize()} " f"system definition.",
    )


OPT_RAW_OUTPUT = typer.Option(
    False,
    "--raw-output",
    "-r",
    show_default=False,
    help="Show information as json formatted string, "
    "use together with --raw-indent option to "
    "pretty print.",
)
OPT_RAW_INDENT = typer.Option(
    None,
    "--raw-indent",
    metavar="NUM_SPACES",
    help="Print json with NUM_SPACES " "spaces of indentation.",
)



from typing import Optional

import typer

from pyrrowhead.management import authorization
from pyrrowhead import rich_console

auth_app = typer.Typer(name="authorization")


@auth_app.command(name="list")
def list_authorization_cli(
    service_definition: Optional[str] = typer.Option(
        None, metavar="SERVICE_DEFINITION"
    ),
    provider_id: Optional[int] = typer.Option(None),
    provider_name: Optional[str] = typer.Option(None),
    consumer_id: Optional[int] = typer.Option(None),
    consumer_name: Optional[str] = typer.Option(None),
):
    """
    Prints all orchestration rules, no filters or sorting options are implemented yet.
    """
    response_data, status = authorization.list_authorization_rules()

    rich_console.print(authorization.create_authorization_table(response_data))


@auth_app.command(name="add")
def add_authorization_cli(
    consumer_id: int = typer.Option(...),
    provider_id: int = typer.Option(...),
    interface_id: int = typer.Option(...),
    service_definition_id: int = typer.Option(...),
):
    """
    Add authorization rule by ids.
    """
    authorization.add_authorization_rule(
        consumer_id, provider_id, interface_id, service_definition_id
    )


@auth_app.command(name="remove")
def remove_authorization_cli():
    """Not implemented."""
    rich_console.print("Not implemented.")
    raise typer.Exit(-1)
    authorization.remove_authorization_rule()



import subprocess
from pathlib import Path
from ipaddress import ip_address
from datetime import datetime, timedelta
from typing import Tuple, List, Optional, Dict, NamedTuple

import yaml
from cryptography.hazmat.primitives.asymmetric import rsa
from cryptography.hazmat.primitives.asymmetric.rsa import RSAPrivateKey
from cryptography.hazmat.primitives import hashes, serialization
from cryptography.hazmat.primitives.serialization.pkcs12 import (
    serialize_key_and_certificates as serialize_p12,
    load_key_and_certificates as load_p12,
)
from cryptography import x509
from cryptography.x509 import Certificate, CertificateSigningRequest
from cryptography.x509.oid import NameOID

from pyrrowhead.types_ import ConfigDict
from pyrrowhead.utils import PyrrowheadError


class KeyCertPair(NamedTuple):
    key: RSAPrivateKey
    cert: Certificate


def set_password_encryption(password: Optional[str] = None):
    if not password:
        return serialization.NoEncryption()

    return serialization.BestAvailableEncryption(password.encode())


def generate_private_key() -> RSAPrivateKey:
    return rsa.generate_private_key(
        public_exponent=65537,
        key_size=2048,
    )


def get_general_name(san_with_prefix: str) -> x509.GeneralName:
    if san_with_prefix.startswith("ip:"):
        return x509.IPAddress(ip_address(san_with_prefix[3:]))
    elif san_with_prefix.startswith("dns:"):
        return x509.DNSName(san_with_prefix[4:])

    raise PyrrowheadError(
        "Pyrrowhead only recognizes IPs prefixed with 'ip:' "
        "or dns strings prefixed with 'dns:'"
        "as valid subject alternative names."
    )


def generate_root_certificate() -> KeyCertPair:
    root_key_alias = "arrowhead.eu"

    root_key = generate_private_key()

    subject = issuer = x509.Name(
        [x509.NameAttribute(NameOID.COMMON_NAME, root_key_alias)]
    )
    root_cert = (
        x509.CertificateBuilder()
        .subject_name(subject)
        .issuer_name(issuer)
        .public_key(root_key.public_key())
        .serial_number(x509.random_serial_number())
        .not_valid_before(datetime.utcnow())
        .not_valid_after(datetime.utcnow() + timedelta(days=365.25 * 10))
        .add_extension(
            x509.BasicConstraints(ca=True, path_length=None),
            critical=False,
        )
        .add_extension(
            x509.SubjectKeyIdentifier.from_public_key(root_key.public_key()),
            critical=False,
        )
        .add_extension(
            x509.SubjectAlternativeName([x509.DNSName(root_key_alias)]), critical=False
        )
        .sign(root_key, hashes.SHA256())
    )

    return KeyCertPair(root_key, root_cert)


def generate_ca_signing_request(
    common_name: str,
    ca: bool,
    path_length: Optional[int],
) -> Tuple[RSAPrivateKey, CertificateSigningRequest]:
    key = generate_private_key()

    csr = (
        x509.CertificateSigningRequestBuilder()
        .subject_name(x509.Name([x509.NameAttribute(NameOID.COMMON_NAME, common_name)]))
        .add_extension(
            x509.BasicConstraints(
                ca=ca,
                path_length=path_length,
            ),
            critical=False,
        )
        .add_extension(
            x509.SubjectAlternativeName([x509.DNSName(common_name)]), critical=False
        )
        .sign(key, hashes.SHA256())
    )

    return key, csr


def generate_system_signing_request(
    common_name: str,
    ip: Optional[str],
    sans: Optional[List[str]],
) -> Tuple[RSAPrivateKey, CertificateSigningRequest]:
    key = generate_private_key()

    general_names: List[x509.GeneralName] = []
    if ip is not None:
        general_names.append(x509.IPAddress(ip_address(ip)))
    if sans is not None:
        general_names.extend(get_general_name(san) for san in sans)

    csr_builder = (
        x509.CertificateSigningRequestBuilder()
        .subject_name(x509.Name([x509.NameAttribute(NameOID.COMMON_NAME, common_name)]))
        .add_extension(
            x509.SubjectAlternativeName(general_names),
            critical=False,
        )
    )

    csr = csr_builder.sign(key, hashes.SHA256())
    return key, csr


def sign_certificate_request(
    csr: CertificateSigningRequest,
    issuer_cert: Certificate,
    issuer_key: RSAPrivateKey,
) -> Certificate:
    cert_builder = (
        x509.CertificateBuilder()
        .subject_name(csr.subject)
        .issuer_name(issuer_cert.subject)
        .public_key(csr.public_key())
        .serial_number(x509.random_serial_number())
        .not_valid_before(datetime.utcnow())
        .not_valid_after(datetime.utcnow() + timedelta(days=365.25 * 10))
        .add_extension(
            x509.SubjectKeyIdentifier.from_public_key(csr.public_key()),
            critical=False,
        )
        .add_extension(
            x509.AuthorityKeyIdentifier.from_issuer_public_key(issuer_key.public_key()),
            critical=False,
        )
    )

    for extension in csr.extensions:
        cert_builder = cert_builder.add_extension(
            extension.value,
            extension.critical,
        )

    cert = cert_builder.sign(issuer_key, hashes.SHA256())

    return cert


def generate_ca_cert(
    common_name: str,
    ca: bool,
    path_length: Optional[int],
    issuer_cert: Certificate,
    issuer_key: RSAPrivateKey,
) -> KeyCertPair:
    ca_key, ca_csr = generate_ca_signing_request(common_name, ca, path_length)
    ca_cert = sign_certificate_request(ca_csr, issuer_cert, issuer_key)

    return KeyCertPair(ca_key, ca_cert)


def generate_system_cert(
    common_name: str,
    ip: Optional[str],
    issuer_cert: Certificate,
    issuer_key: RSAPrivateKey,
    sans: Optional[List[str]] = None,
) -> KeyCertPair:
    system_key, system_csr = generate_system_signing_request(common_name, ip, sans)
    system_cert = sign_certificate_request(system_csr, issuer_cert, issuer_key)

    return KeyCertPair(system_key, system_cert)


def generate_core_system_certs(
    cloud_config: ConfigDict, cloud_cert, cloud_key
) -> Dict[str, KeyCertPair]:
    cloud_name = cloud_config["cloud"]["cloud_name"]
    org_name = cloud_config["cloud"]["organization_name"]
    return {
        core_system["system_name"]: generate_system_cert(
            common_name=f'{core_system["system_name"]}.'
            f"{cloud_name}.{org_name}.arrowhead.eu",
            ip=core_system["address"],
            issuer_cert=cloud_cert,
            issuer_key=cloud_key,
            sans=cloud_config["cloud"]["core_san"],
        )
        for core_system in cloud_config["cloud"]["core_systems"].values()
    }


def generate_client_system_certs(
    cloud_config: ConfigDict, cloud_cert, cloud_key
) -> Dict[str, KeyCertPair]:
    cloud_name = cloud_config["cloud"]["cloud_name"]
    org_name = cloud_config["cloud"]["organization_name"]
    return {
        client_id: generate_system_cert(
            f'{client_system["system_name"]}.{cloud_name}.{org_name}.arrowhead.eu',
            client_system["address"],
            cloud_cert,
            cloud_key,
            client_system.get("sans", None),
        )
        for client_id, client_system in cloud_config["cloud"]["client_systems"].items()
    }


def store_system_files(
    core_keycert: KeyCertPair,
    core_cert_path: Path,
    root_cert,
    org_cert,
    cloud_cert,
    password: Optional[str],
):
    core_name = core_keycert.cert.subject.get_attributes_for_oid(NameOID.COMMON_NAME)[
        0
    ].value

    with open(core_cert_path.with_suffix(".p12"), "wb") as p12_file:
        p12_file.write(
            serialize_p12(
                name=core_name.encode(),
                key=core_keycert.key,
                cert=core_keycert.cert,
                cas=[cloud_cert, org_cert, root_cert],
                encryption_algorithm=set_password_encryption(password),
            )
        )
        print(root_cert.subject.get_attributes_for_oid(NameOID.COMMON_NAME))
    with open(core_cert_path.with_suffix(".crt"), "wb") as crt_file:
        crt_file.write(core_keycert.cert.public_bytes(serialization.Encoding.PEM))
    with open(core_cert_path.with_suffix(".key"), "wb") as key_file:
        key_file.write(
            core_keycert.key.private_bytes(
                serialization.Encoding.PEM,
                serialization.PrivateFormat.TraditionalOpenSSL,
                encryption_algorithm=set_password_encryption(password),
            )
        )


def store_sysop(
    sysop_certs: KeyCertPair,
    root_cert: Certificate,
    org_cert: Certificate,
    cloud_cert: Certificate,
    cert_directory: Path,
    password: Optional[str],
):
    with open(sysop_p12 := cert_directory / "sysop.p12", "wb") as p12_file:
        cloud_name = cloud_cert.subject.get_attributes_for_oid(NameOID.COMMON_NAME)[
            0
        ].value
        p12_file.write(
            serialize_p12(
                name=f"sysop.{cloud_name}".encode(),
                key=sysop_certs.key,
                cert=sysop_certs.cert,
                cas=[cloud_cert, org_cert, root_cert],
                encryption_algorithm=set_password_encryption(password),
            )
        )
    with open(sysop_p12.with_suffix(".crt"), "wb") as crt_file:
        crt_file.write(sysop_certs.cert.public_bytes(serialization.Encoding.PEM))
    with open(sysop_p12.with_suffix(".key"), "wb") as key_file:
        key_file.write(
            sysop_certs.key.private_bytes(
                serialization.Encoding.PEM,
                serialization.PrivateFormat.TraditionalOpenSSL,
                encryption_algorithm=serialization.NoEncryption(),
            )
        )
    with open(sysop_p12.with_suffix(".ca"), "wb") as ca_file:
        ca_file.write(cloud_cert.public_bytes(serialization.Encoding.PEM))
        ca_file.write(org_cert.public_bytes(serialization.Encoding.PEM))
        ca_file.write(root_cert.public_bytes(serialization.Encoding.PEM))


def store_truststore(cloud_cert: Certificate, cert_directory: Path, password: str):
    cloud_long_name = cloud_cert.subject.get_attributes_for_oid(NameOID.COMMON_NAME)[
        0
    ].value
    cloud_short_name, *_ = cloud_long_name.split(".")
    subprocess.run(
        f"keytool -importcert -trustcacerts"
        f" -noprompt -storepass {password!s}"
        f" -keystore {cert_directory}/truststore.p12"
        f" -file {cert_directory}/{cloud_short_name}.crt"
        f" -alias {cloud_long_name}".split(),
        capture_output=True,
    )


def generate_and_store_root_files(
    root_cert_directory: Path, password=Optional[str]
) -> KeyCertPair:
    root_keycert = generate_root_certificate()
    with open(root_cert_directory / "root.p12", "wb") as root_p12:
        root_p12.write(
            serialize_p12(
                name=b"arrowhead.eu",
                key=root_keycert.key,
                cert=root_keycert.cert,
                cas=None,
                encryption_algorithm=set_password_encryption(password),
            )
        )
    with open(root_cert_directory / "root.crt", "wb") as root_crt:
        root_crt.write(root_keycert.cert.public_bytes(serialization.Encoding.PEM))

    return root_keycert


def generate_and_store_org_files(
    org_name: str,
    org_cert_dir: Path,
    root_key,
    root_cert,
    password: Optional[str],
) -> KeyCertPair:
    org_keycert = generate_ca_cert(
        f"{org_name}.arrowhead.eu",
        ca=True,
        path_length=None,
        issuer_cert=root_cert,
        issuer_key=root_key,
    )
    with open(org_cert_dir / f"{org_name}.p12", "wb") as org_p12:
        org_p12.write(
            serialize_p12(
                name=f"{org_name}.arrowhead.eu".encode(),
                key=org_keycert.key,
                cert=org_keycert.cert,
                cas=[root_cert],
                encryption_algorithm=set_password_encryption(password),
            )
        )
    with open(org_cert_dir / f"{org_name}.crt", "wb") as crt_org:
        crt_org.write(org_keycert.cert.public_bytes(serialization.Encoding.PEM))
    with open(org_cert_dir / f"{org_name}.key", "wb") as key_org:
        key_org.write(
            org_keycert.key.private_bytes(
                serialization.Encoding.PEM,
                serialization.PrivateFormat.TraditionalOpenSSL,
                encryption_algorithm=serialization.NoEncryption(),
            )
        )

    return org_keycert


def generate_and_store_cloud_cert(
    cloud_name: str,
    org_name: str,
    cloud_cert_dir: Path,
    org_key: RSAPrivateKey,
    org_cert: Certificate,
    root_cert: Certificate,
    password: Optional[str],
) -> KeyCertPair:
    cloud_keycert = generate_ca_cert(
        f"{cloud_name}.{org_name}.arrowhead.eu",
        ca=True,
        path_length=2,
        issuer_cert=org_cert,
        issuer_key=org_key,
    )
    with open(cloud_cert_dir / f"{cloud_name}.p12", "wb") as cloud_p12:
        cloud_p12.write(
            serialize_p12(
                name=f"{cloud_name}.{org_name}.arrowhead.eu".encode(),
                key=cloud_keycert.key,
                cert=cloud_keycert.cert,
                cas=[org_cert, root_cert],
                encryption_algorithm=set_password_encryption(password),
            )
        )
    with open(cloud_cert_dir / f"{cloud_name}.crt", "wb") as cloud_crt:
        cloud_crt.write(cloud_keycert.cert.public_bytes(serialization.Encoding.PEM))
    with open(cloud_cert_dir / f"{cloud_name}.key", "wb") as cloud_key_file:
        cloud_key_file.write(
            cloud_keycert.key.private_bytes(
                serialization.Encoding.PEM,
                serialization.PrivateFormat.TraditionalOpenSSL,
                encryption_algorithm=serialization.NoEncryption(),
            )
        )
    return cloud_keycert


def generate_cloud_files(
    cloud_config: ConfigDict,
    cloud_cert_dir: Path,
    cloud_key: RSAPrivateKey,
    cloud_cert: Certificate,
    org_cert: Certificate,
    root_cert: Certificate,
    password: str,
):
    cloud_name = cloud_config["cloud"]["cloud_name"]
    org_name = cloud_config["cloud"]["organization_name"]

    core_system_certs = generate_core_system_certs(cloud_config, cloud_cert, cloud_key)
    client_system_certs = generate_client_system_certs(
        cloud_config, cloud_cert, cloud_key
    )
    sysop_cert = generate_system_cert(
        f"sysop.{cloud_name}.{org_name}.arrowhead.eu",
        None,
        issuer_cert=cloud_cert,
        issuer_key=cloud_key,
    )

    if not all(
        (cloud_cert_dir / "sysop").with_suffix(suf).exists()
        for suf in (".p12", ".crt", ".key", ".ca")
    ):
        store_sysop(
            sysop_cert, root_cert, org_cert, cloud_cert, cloud_cert_dir, password
        )
    for core_name, core_cert in core_system_certs.items():
        if all(
            (core_cert_path := cloud_cert_dir / core_name).with_suffix(suf).exists()
            for suf in (".p12", ".crt", ".key")
        ):
            continue
        store_system_files(
            core_cert,
            core_cert_path,  # noqa
            root_cert,
            org_cert,
            cloud_cert,
            password,
        )
    for client_name, client_cert in client_system_certs.items():
        if all(
            (client_cert_path := cloud_cert_dir / client_name).with_suffix(suf).exists()
            for suf in {".p12", ".crt", ".key"}
        ):
            continue
        store_system_files(
            client_cert,
            client_cert_path,  # noqa
            root_cert,
            org_cert,
            cloud_cert,
            password,
        )
    store_truststore(cloud_cert, cloud_cert_dir, password)


def setup_certificates(cloud_config_path: Path, password: str):
    with open(cloud_config_path, "r") as cloud_config_file:
        cloud_config: ConfigDict = yaml.safe_load(cloud_config_file)

    cloud_name = cloud_config["cloud"]["cloud_name"]
    org_name = cloud_config["cloud"]["organization_name"]

    cloud_dir = cloud_config_path.parent
    cloud_cert_dir = cloud_dir / "certs/crypto/"
    org_cert_dir = cloud_dir.parent / "org-certs/crypto/"
    root_cert_dir = cloud_dir.parent / "root-certs/crypto"

    # TODO: Invert this horrible stack of ifs
    if not cloud_cert_dir.exists():
        print("Cloud cert does not exist, generating from organization cert...")
        if not org_cert_dir.exists():
            print("Organization cert does not exist, generating from root cert...")
            if not root_cert_dir.exists():
                print("Root cert does not exist, generating self-signed root cert.")
                root_cert_dir.mkdir(parents=True)
                root_key, root_cert = generate_and_store_root_files(
                    root_cert_dir, password
                )
            else:
                with open(root_cert_dir / "root.p12", "rb") as root_p12:
                    root_key, root_cert, *_ = load_p12(  # type: ignore
                        root_p12.read(), password.encode()
                    )  # noqa
                    if not isinstance(root_key, RSAPrivateKey) or not isinstance(
                        root_cert, Certificate
                    ):
                        raise PyrrowheadError("Could not open root key or certificate")

            org_cert_dir.mkdir(parents=True)
            org_key, org_cert = generate_and_store_org_files(  # type: ignore
                org_name,
                org_cert_dir,
                root_key,
                root_cert,
                password,
            )
        else:
            with open(org_cert_dir / f"{org_name}.p12", "rb") as org_p12:
                org_key, org_cert, ca_certs = load_p12(  # type: ignore
                    org_p12.read(), password.encode()
                )
                if not isinstance(org_key, RSAPrivateKey) or not isinstance(
                    org_cert, Certificate
                ):
                    raise PyrrowheadError("Could not read org key or certificate")
                if len(ca_certs) != 1:
                    raise PyrrowheadError(
                        f"Organization certificate can only have one CA, "
                        f"currently has {len(ca_certs)}."
                    )
                root_cert = ca_certs[0]

        cloud_cert_dir.mkdir(parents=True)
        cloud_key, cloud_cert = generate_and_store_cloud_cert(
            cloud_name,
            org_name,
            cloud_cert_dir,
            org_key,
            org_cert,
            root_cert,
            password,
        )
    else:
        with open(cloud_cert_dir / f"{cloud_name}.p12", "rb") as cloud_cert_file:
            cloud_key, cloud_cert, ca_certs = load_p12(  # type: ignore
                cloud_cert_file.read(), password.encode()
            )
            if len(ca_certs) != 2:
                raise RuntimeError(
                    f"Cloud cert must have exactly two CAs, "
                    f"currently have {len(ca_certs)}."
                )
            org_cert, root_cert = ca_certs

    generate_cloud_files(
        cloud_config,
        cloud_cert_dir,
        cloud_key,
        cloud_cert,
        org_cert,
        root_cert,
        password,
    )



from typing import Optional
from pathlib import Path

from rich import box
from rich.table import Table, Column

from ..management.utils import get_service, post_service, delete_service
from ..utils import get_core_system_address_and_port, get_active_cloud_directory


def list_systems():
    active_cloud_directory = get_active_cloud_directory()
    address, port, secure, scheme = get_core_system_address_and_port(
        "service_registry",
        active_cloud_directory,
    )
    scheme = "https" if secure else "http"
    response = get_service(
        f"{scheme}://{address}:{port}/serviceregistry/mgmt/systems",
        active_cloud_directory,
    )
    response_data, status = response.json(), response.status_code
    return response_data, status


def add_system(
    system_name: str,
    system_address: str,
    system_port: int,
    certificate_file: Optional[Path] = None,
):
    active_cloud_directory = get_active_cloud_directory()
    address, port, secure, scheme = get_core_system_address_and_port(
        "service_registry",
        active_cloud_directory,
    )

    system_record = {
        "systemName": system_name,
        "address": system_address,
        "port": system_port,
    }

    response_data = post_service(
        f"{scheme}://{address}:{port}/serviceregistry/mgmt/systems",
        active_cloud_directory,
        json=system_record,
    ).json()

    return response_data


def remove_system(system_id: int):
    active_cloud_directory = get_active_cloud_directory()
    address, port, secure, scheme = get_core_system_address_and_port(
        "service_registry",
        active_cloud_directory,
    )

    response = delete_service(
        f"{scheme}://{address}:{port}/serviceregistry/mgmt/systems/{system_id}",
        active_cloud_directory,
    )

    return response.json(), response.status_code


def create_system_table(response):
    system_data = response

    system_table = Table(
        Column(header="id", style="red"),
        Column(header="System name", style="blue"),
        Column(header="Address", style="purple"),
        Column(header="Port", style="bright_white"),
        title="Registered systems",
        box=box.SIMPLE,
    )

    for system in system_data["data"]:
        system_table.add_row(
            str(system["id"]),
            system["systemName"],
            system["address"],
            str(system["port"]),
        )

    return system_table



from pathlib import Path
from typing import Optional, List, Tuple

import typer

from pyrrowhead import rich_console
from pyrrowhead.cloud.installation import install_cloud, uninstall_cloud
from pyrrowhead.cloud.create import CloudConfiguration, create_cloud_config
from pyrrowhead.cloud.start import start_local_cloud
from pyrrowhead.cloud.stop import stop_local_cloud
from pyrrowhead.cloud.configuration import enable_ssl as enable_ssl_func
from pyrrowhead.cloud.client_add import add_client_system
from pyrrowhead.utils import (
    switch_directory,
    set_active_cloud as set_active_cloud_func,
    get_config,
    check_valid_dns,
    PyrrowheadError,
)
from pyrrowhead.constants import (
    OPT_CLOUDS_DIRECTORY,
    OPT_CLOUD_NAME,
    OPT_ORG_NAME,
    ARG_CLOUD_IDENTIFIER,
)

cloud_app = typer.Typer(
    name="cloud",
    help="Used to set up, configure, start, and stop local clouds using "
    "the appropriate subcommand. See list below.",
)


def decide_cloud_directory(
    cloud_identifier: Optional[str],
    cloud_name: Optional[str],
    organization_name: Optional[str],
    clouds_directory: Path,
) -> Tuple[Path, str]:
    if (
        isinstance(cloud_identifier, str)
        and len(split_cloud_identifier := cloud_identifier.split(".")) == 2
    ):
        ret = (
            clouds_directory.joinpath(
                *[part for part in reversed(split_cloud_identifier)]
            ),
            cloud_identifier,
        )
    elif cloud_name is not None and organization_name is not None:
        ret = (
            clouds_directory.joinpath(organization_name, cloud_name),
            f"{cloud_name}.{organization_name}",
        )
    elif isinstance(cloud_identifier, str) and cloud_identifier != "":
        ret = (clouds_directory, cloud_identifier)
    else:
        rich_console.print("Could not decide local cloud.")
        raise typer.Exit(-1)

    if not ret[0].exists():
        rich_console.print(f'Could not find local cloud "{cloud_identifier}"')
        raise typer.Exit(-1)

    return ret


@cloud_app.command()
def configure(
    cloud_identifier: str = ARG_CLOUD_IDENTIFIER,
    cloud_name: Optional[str] = OPT_CLOUD_NAME,
    organization_name: Optional[str] = OPT_ORG_NAME,
    clouds_directory: Path = OPT_CLOUDS_DIRECTORY,
    enable_ssl: Optional[bool] = typer.Option(None, "--enable-ssl/--disable-ssl"),
):
    """
    Statically configures an existing local cloud.

    The local cloud needs to be reinstalled after being configured
    to make sure the changes are applied.
    """
    target, cloud_identifier = decide_cloud_directory(
        cloud_identifier,
        cloud_name,
        organization_name,
        clouds_directory,
    )
    with switch_directory(target):
        if enable_ssl is not None:
            enable_ssl_func(enable_ssl)


@cloud_app.command()
def list(
    # organization_filter: str = typer.Option('', '--organization', '-o'),
):
    """
    Lists all local clouds.
    """
    config = get_config()

    for cloud_identifier, directory in config["local-clouds"].items():
        if not Path(directory).exists():
            rich_console.print(cloud_identifier, "Path does not exist", style="red")
        else:
            rich_console.print(cloud_identifier, directory)


@cloud_app.command()
def install(
    cloud_identifier: str = ARG_CLOUD_IDENTIFIER,
    cloud_name: Optional[str] = OPT_CLOUD_NAME,
    organization_name: Optional[str] = OPT_ORG_NAME,
    cloud_directory: Path = OPT_CLOUDS_DIRECTORY,
):
    """
    Installs cloud by creating certificate files and core service configuration files.

    CLOUD_NAME and ORG_name are the cloud and organization names used in the generated certificates.
    """  # noqa
    target, cloud_identifier = decide_cloud_directory(
        cloud_identifier,
        cloud_name,
        organization_name,
        cloud_directory,
    )

    config_file = target / "cloud_config.yaml"

    if not target.exists():
        raise RuntimeError(
            "Target cloud is not set up properly,"
            " run `pyrrowhead cloud setup` before installing cloud."
        )

    install_cloud(config_file, target)


@cloud_app.command()
def uninstall(
    cloud_identifier: str = ARG_CLOUD_IDENTIFIER,
    cloud_name: Optional[str] = OPT_CLOUD_NAME,
    organization_name: Optional[str] = OPT_ORG_NAME,
    clouds_directory: Path = OPT_CLOUDS_DIRECTORY,
    complete: bool = typer.Option(
        False,
        "--complete",
        help="Completely removes all files, including cloud_config.yaml",
    ),
    # keep_root: bool = typer.Option(False, '--keep-root', ),
):
    """
    Uninstalls cloud by removing certificate files and core service configuration files.
    Keeps the cloud_config.yaml file by default.

    CLOUD_NAME and ORG_name are the cloud and organization names used in the generated certificates.
    """  # noqa
    target, cloud_identifier = decide_cloud_directory(
        cloud_identifier,
        cloud_name,
        organization_name,
        clouds_directory,
    )

    stop_local_cloud(target)
    uninstall_cloud(target, complete)


@cloud_app.command()
def create(
    cloud_identifier: Optional[str] = ARG_CLOUD_IDENTIFIER,
    cloud_name: Optional[str] = OPT_CLOUD_NAME,
    organization_name: Optional[str] = OPT_ORG_NAME,
    installation_target: Path = OPT_CLOUDS_DIRECTORY,
    ip_network: str = typer.Option(
        "172.16.1.0/24",
        metavar="IP",
        help="IP network the docker network uses to run the local clouds",
    ),
    core_san: Optional[List[str]] = typer.Option(
        None,
        "--san",
        metavar="SUBJECT_ALTERNATIVE_NAME",
        help="Subject alternative names to include in the core system certificates."
        " An example is the IP-address of the device the core systems are running"
        " on. IPs should be prefixed with 'ip:' and DNS strings prefixed "
        "with 'dns:', for example ip:127.0.0.1 and dns:host123.example-org.com",
    ),
    ssl_enabled: Optional[bool] = typer.Option(
        True,
        "--ssl-enabled/--ssl-disabled",
        show_default=False,
        help="Enabled/disable local cloud security. Enabled by default. "
        "SSL rarely be disabled.",
    ),
    do_install: bool = typer.Option(
        False, "--install", help="Install local cloud after running the setup command."
    ),
    include: Optional[List[CloudConfiguration]] = typer.Option(
        [],
        case_sensitive=False,
        help="Core systems to include apart from the mandatory core systems. "
        "--include eventhandler includes the eventhandler. "
        "--include intercloud includes the gateway and gatekeeper. "
        "--include onboarding includes the system and device registry, "
        "certificate authority and onboarding systems.",
    ),
):
    """
    Sets up local clouds by creating a folder structure and cloud_config.yaml file.

    CLOUD_NAME and ORG_name are the cloud and organization names used in the generated certificates.
    """  # noqa
    if not cloud_identifier and not cloud_name and not organization_name:
        rich_console.print(
            "You must either provide the 'CLOUD_IDENTIFIER' argument or both the "
            "'CLOUD_NAME' and 'ORG_NAME' options."
        )
        raise typer.Exit(-1)
    if (
        cloud_identifier is None
        and cloud_name is not None
        and organization_name is not None
    ):
        cloud_identifier = ".".join((cloud_name, organization_name))
    elif cloud_identifier is not None:
        cloud_name, organization_name = cloud_identifier.split(".")
    else:
        rich_console.print(
            "CLOUD_IDENTIFIER, or CLOUD_NAME or ORG_NAME are of unknown types."
        )
        raise typer.Exit(-1)

    if not check_valid_dns(cloud_identifier):
        rich_console.print(f'"{cloud_identifier}" is not a valid cloud identifier')
        raise typer.Exit(-1)

    try:
        create_cloud_config(
            target_directory=installation_target,
            cloud_identifier=cloud_identifier,
            cloud_name=cloud_name,
            organization_name=organization_name,
            ssl_enabled=ssl_enabled,
            ip_subnet=ip_network,
            core_san=core_san,
            do_install=do_install,
            include=include,
        )
    except PyrrowheadError as e:
        rich_console.print(str(e))
        raise typer.Exit(-1)


@cloud_app.command()
def up(
    cloud_identifier: str = ARG_CLOUD_IDENTIFIER,
    cloud_name: Optional[str] = OPT_CLOUD_NAME,
    organization_name: Optional[str] = OPT_ORG_NAME,
    clouds_directory: Path = OPT_CLOUDS_DIRECTORY,
    set_active_cloud: bool = typer.Option(
        True,
        " /--no-set-active",
        " /-N",
        show_default=False,
        help="Does not set this cloud as the active cloud, "
        "useful if you want to start another cloud in the background.",
    ),
):
    """
    Starts local cloud core system docker containers.

    If this command fails during the mysql startup, it might be because you are running
    another mysql instance on port 3306. You must either terminate that service (e.g. running
    `systemctl stop mysql.service`) or change the port of mysql in the configuration and reinstall
    the local cloud before starting it again.

    This command might take a while if this is the first time starting a local cloud on this machine
    as docker needs to pull the images.
    """  # noqa
    target, cloud_identifier = decide_cloud_directory(
        cloud_identifier,
        cloud_name,
        organization_name,
        clouds_directory,
    )
    try:
        start_local_cloud(target)
        if set_active_cloud:
            set_active_cloud_func(cloud_identifier)
    except KeyboardInterrupt:
        stop_local_cloud(target)
        raise typer.Abort()


@cloud_app.command()
def down(
    cloud_identifier: str = ARG_CLOUD_IDENTIFIER,
    cloud_name: Optional[str] = OPT_CLOUD_NAME,
    organization_name: Optional[str] = OPT_ORG_NAME,
    clouds_directory: Path = OPT_CLOUDS_DIRECTORY,
):
    """
    Shuts down local cloud.
    """
    target, cloud_identifier = decide_cloud_directory(
        cloud_identifier,
        cloud_name,
        organization_name,
        clouds_directory,
    )
    stop_local_cloud(target)
    set_active_cloud_func("")


@cloud_app.command(name="client-add")
def system(
    cloud_identifier: str = ARG_CLOUD_IDENTIFIER,
    cloud_name: Optional[str] = OPT_CLOUD_NAME,
    organization_name: Optional[str] = OPT_ORG_NAME,
    clouds_directory: Path = OPT_CLOUDS_DIRECTORY,
    system_name: str = typer.Option(
        ..., "--name", "-n", metavar="SYSTEM_NAME", help="System name"
    ),
    system_address: Optional[str] = typer.Option(
        None, "--addr", "-a", metavar="ADDRESS", help="System address"
    ),
    system_port: Optional[int] = typer.Option(
        None, "--port", "-p", metavar="PORT", help="System port"
    ),
    system_addl_addr: Optional[List[str]] = typer.Option(
        None,
        "--san",
        "-s",
        metavar="SAN",
        help="Client subject alternative name.",
    ),
):
    """
    Adds system to the cloud configuration.
    """
    target, cloud_identifier = decide_cloud_directory(
        cloud_identifier,
        cloud_name,
        organization_name,
        clouds_directory,
    )

    config_file = target / "cloud_config.yaml"

    try:
        add_client_system(
            config_file,
            system_name,
            system_address,
            system_port,
            system_addl_addr,
        )
    except PyrrowheadError as e:
        rich_console.print(str(e))
        raise typer.Exit(-1)



from pyrrowhead.management.cli.service import sr_app  # noqa
from pyrrowhead.management.cli.orchestration import orch_app  # noqa
from pyrrowhead.management.cli.authorization import auth_app  # noqa
from pyrrowhead.management.cli.system import sys_app  # noqa



db_passwords = {
    "service_registry": "ZzNNpxrbZGVvfJ8",
    "authorization": "hqZFUkuHxhekio3",
    "orchestrator": "KbgD2mTr8DQ4vtc",
    "event_handler": "gRLjXbqu9YwYhfK",
    "gateway": "LfiSM9DpGfDEP5g",
    "gatekeeper": "fbJKYzKhU5t8QtT",
    "system_registry": "Kh12Hhgaxzo7haf",
    "device_registry": "iooHU87hNGUalht",
    "onboarding_controller": "JKgh1as5f6oi7aV",
    "certificate_authority": "FsdG6Kgf9QpPfv2",
}



import json
from pathlib import Path
from typing import Optional

import typer
from rich.syntax import Syntax

from pyrrowhead.management import systemregistry
from pyrrowhead import rich_console

sys_app = typer.Typer(name="systems")


@sys_app.command(name="list")
def list_systems_cli(
    raw_output: bool = typer.Option(False, "--raw-output", "-r", show_default=False),
    indent: Optional[int] = typer.Option(None, "--raw-indent"),
):
    """List systems registered in the local cloud"""
    response_data, status = systemregistry.list_systems()

    if raw_output:
        if status >= 400:
            rich_console.print(f"Error code {status}.")
        rich_console.print(Syntax(json.dumps(response_data, indent=indent), "json"))
        raise typer.Exit()

    table = systemregistry.create_system_table(response_data)

    rich_console.print(table)


@sys_app.command(name="add")
def add_system_cli(
    system_name: str,
    # Add a callback to verify ip
    system_address: str = typer.Argument(..., metavar="ADDRESS"),
    system_port: int = typer.Argument(..., metavar="PORT"),
    certificate_file: Optional[Path] = None,
):
    response_data = systemregistry.add_system(
        system_name, system_address, system_port, certificate_file
    )

    rich_console.print(Syntax(json.dumps(response_data, indent=2), "json"))


@sys_app.command(name="remove")
def remove_system_cli(system_id: int):
    """Remove system by id."""
    response_data, status = systemregistry.remove_system(system_id)



import json
from typing import Optional, Tuple

import typer
from rich.syntax import Syntax
from rich.text import Text

from pyrrowhead.management import common, serviceregistry
from pyrrowhead.management.common import AccessPolicy
from pyrrowhead import rich_console

sr_app = typer.Typer(
    name="services", help="Service related commands. See list for further information."
)


@sr_app.command(name="list")
def services_list_cli(
    service_definition: Optional[str] = typer.Option(
        None,
        show_default=False,
        metavar="SERVICE_DEFINITION",
        help="Filter services by SERVICE_DEFINITION",
    ),
    system_name: Optional[str] = typer.Option(
        None,
        show_default=False,
        metavar="SYSTEM_NAME",
        help="Filter services by SYSTEM_NAME",
    ),
    system_id: Optional[int] = typer.Option(
        None,
        show_default=False,
        metavar="SYSTEM_ID",
        help="Filter services by SYSTEM_ID",
    ),
    show_service_uri: bool = typer.Option(
        False, "--show-service-uri", "-u", show_default=False, help="Show service uri"
    ),
    show_access_policy: bool = typer.Option(
        False,
        "--show-access-policy",
        "-c",
        show_default=False,
        help="Show access policy",
    ),
    show_provider: bool = typer.Option(
        None, "--show-provider", "-s", help="Show provider system"
    ),
    raw_output: bool = common.OPT_RAW_OUTPUT,
    indent: Optional[int] = common.OPT_RAW_INDENT,
):
    """
    List services registered in the active local cloud, sorted by ID.

    Services shown can be filtered by service definition or system. More information about the
    services can be seen with the -usc flags. The raw json data is accessed by the -r flag.
    """  # noqa
    exclusive_options = (service_definition, system_name, system_id)
    if len(list(option for option in exclusive_options if option is not None)) > 1:
        raise RuntimeError(
            "Only one of the options <--service-definition, --system-name, --system-id>"
            " may be used."
        )

    try:
        list_data = serviceregistry.list_services(
            service_definition,
            system_name,
            system_id,
        )
    except RuntimeError as e:
        rich_console.print(e)
        raise typer.Exit(code=-1)

    if raw_output:
        rich_console.print(Syntax(json.dumps(list_data, indent=indent), "json"))
        raise typer.Exit()

    service_table = serviceregistry.create_service_table(
        list_data, show_provider, show_access_policy, show_service_uri
    )

    rich_console.print(service_table)


@sr_app.command(name="inspect")
def inspect_service_cli(
    service_id: int = typer.Argument(
        ..., metavar="SERVICE_ID", help="ID of service to inspect."
    ),
    raw_output: Optional[bool] = common.OPT_RAW_OUTPUT,
    raw_indent: Optional[int] = common.OPT_RAW_INDENT,
):
    """
    Show information about service given by ID.
    """
    response_data, status = serviceregistry.inspect_service(service_id)

    if status >= 400:
        rich_console.print(
            f"Error occured when trying to inspect service with id {service_id} due to:"
            f' {response_data["exceptionType"]}, {response_data["errorMessage"]}'
        )
        raise typer.Exit(-1)

    if raw_output:
        rich_console.print(Syntax(json.dumps(response_data, indent=raw_indent), "json"))
        raise typer.Exit()

    serviceregistry.render_service(response_data)


@sr_app.command(name="add")
def add_service_cli(
    service_definition: str = common.ARG_SERVICE_DEFINITION,
    uri: str = common.ARG_SERVICE_URI,
    interface: str = common.ARG_SERVICE_INTERFACE,
    access_policy: AccessPolicy = typer.Option(
        AccessPolicy.CERTIFICATE,
        metavar="ACCESS_POLICY",
        help='Must be one of three values: "NOT_SECURE", ' '"CERTIFICATE", or "TOKEN"',
    ),
    system: Tuple[str, str, int] = typer.Option(
        ...,
        show_default=False,
        metavar="SYSTEM_NAME ADDRESS PORT",
        help="Provider system definition.",
    ),
    system_id: Optional[int] = typer.Option(None, help="Not yet supported"),
):
    """
    Add service to service registry
    """
    # TODO: Implement system_id option
    if all((all(system), system_id)):
        rich_console.print("--System and --system-id are mutually exclusive options.")
        raise typer.Exit()
    elif not any((all(system), system_id)):
        rich_console.print("One option of --system or --system-id must be set.")
        raise typer.Exit()

    try:
        response_data, response_code = serviceregistry.add_service(
            service_definition,
            uri,
            interface,
            access_policy,
            system,
        )
    except IOError as e:
        rich_console.print(e)
        raise typer.Exit(-1)

    # Add service code
    if response_code >= 400:
        rich_console.print(
            Text(f'Service registration failed: {response_data["errorMessage"]}')
        )
        raise typer.Exit(-1)

    serviceregistry.render_service(response_data)


@sr_app.command(name="remove")
def remove_service_cli(
    id: int = typer.Argument(..., metavar="SERVICE_ID", help="Id of service to remove"),
):
    """
    Remove service from service registry by ID.
    """
    try:
        response_data, status = serviceregistry.delete_service(id)
    except IOError as e:
        rich_console.print(e)
        raise typer.Exit(-1)

    if status in {400, 401, 500}:
        rich_console.print(
            Text(f'Service unregistration failed: {response_data["errorMessage"]}')
        )
        raise typer.Exit(-1)



import shutil
from pathlib import Path
from typing import Optional

from pyrrowhead.utils import get_local_cloud_directory, PyrrowheadError
from pyrrowhead.constants import ORG_CERT_DIR, ROOT_CERT_DIR
from pyrrowhead.new_certificate_generation.generate_certificates import (
    generate_and_store_root_files,
    generate_and_store_org_files,
    KeyCertPair,
    load_p12,
)


def load_root_certificates(org_dir: Path, password: str):
    root_cert_dir = org_dir / ROOT_CERT_DIR / "crypto"
    with open(root_cert_dir / "root.p12", "rb") as root_p12:
        root_key, root_cert, *_ = load_p12(root_p12.read(), password.encode())
    return root_key, root_cert


def create_root_certificates(org_dir: Path, password: str) -> KeyCertPair:
    root_cert_dir = org_dir / ROOT_CERT_DIR / "crypto"
    root_cert_dir.mkdir(parents=True)
    pair = generate_and_store_root_files(root_cert_dir, password)
    return pair


def create_org_certificates(
    org_name: str,
    org_dir: Path,
    root_key,
    root_cert,
    password,
) -> KeyCertPair:
    org_cert_dir = org_dir / ORG_CERT_DIR / "crypto"
    org_cert_dir.mkdir(parents=True)
    pair = generate_and_store_org_files(
        org_name, org_cert_dir, root_key, root_cert, password
    )
    return pair


def copy_org_certificates(
    org_name: str,
    key_path: Path,
    cert_path: Optional[Path] = None,
):
    org_dir = get_local_cloud_directory() / org_name
    org_cert_dir = org_dir / ORG_CERT_DIR / "crypto"
    shutil.copy(key_path, org_cert_dir / f"{org_name}.key")

    if cert_path is not None:
        shutil.copy(key_path, org_cert_dir / f"{org_name}.crt")


def populate_org_dir(org_name: str, password: str):
    org_dir = get_local_cloud_directory() / org_name

    if not all(
        org_dir.joinpath(subdir).exists() for subdir in (ORG_CERT_DIR, ROOT_CERT_DIR)
    ):
        root_key_cert = create_root_certificates(org_dir, password)
        create_org_certificates(org_name, org_dir, *root_key_cert, password)
    elif not org_dir.joinpath(ORG_CERT_DIR).exists():
        root_key_cert = create_root_certificates(org_dir, password)
        root_key, root_cert = load_root_certificates(org_dir, password)
        create_org_certificates(org_name, org_dir, root_key, root_cert, password)


def mk_org_dir(org_name: str):
    org_dir = get_local_cloud_directory() / org_name

    if org_dir.is_dir():
        raise PyrrowheadError(f'Organization "{org_name}" already exists')
    if not org_dir.exists():
        org_dir.mkdir()



"""
This module is work-in-progress.
"""
from pathlib import Path
from typing import Optional

import typer

from pyrrowhead.org.initialize_org import (
    mk_org_dir,
    populate_org_dir,
    copy_org_certificates,
)
from pyrrowhead import rich_console
from pyrrowhead.utils import (
    PyrrowheadError,
    check_valid_dns,
)
from pyrrowhead.constants import (
    ARG_ORG_NAME,
)


org_app = typer.Typer(name="org")


@org_app.command()
def create(org_name: str = ARG_ORG_NAME):
    """
    Initializes an empty organization with name ORG_NAME.
    """
    if not check_valid_dns(org_name):
        rich_console.print(
            PyrrowheadError(f'"{org_name}" is not a valid organization name.')
        )
        raise typer.Exit(-1)

    try:
        mk_org_dir(org_name)
    except PyrrowheadError as e:
        rich_console.print(str(e))
        raise typer.Exit(-1)

    rich_console.print(f"Created organization {org_name}")


@org_app.command()
def cert_gen(org_name: str = ARG_ORG_NAME):
    password = "123456"
    populate_org_dir(org_name, password)


@org_app.command()
def add_cert(
    org_name: str = ARG_ORG_NAME,
    key_path: Path = typer.Option(..., "--key-path", "-k"),
    cert_path: Optional[Path] = typer.Option(..., "--cert-path", "-c"),
    certgen: Optional[bool] = typer.Option(False, "--certgen", "-g"),
):
    if not key_path.is_file():
        rich_console.print(f"Cannot find key at {key_path}")
        raise typer.Exit(-1)

    if cert_path is not None:
        copy_org_certificates(org_name, key_path, cert_path)


@org_app.command()
def add_p12(org_name: str = ARG_ORG_NAME):
    pass



from pathlib import Path

import typer
import yaml
import yamlloader  # type: ignore

from pyrrowhead import rich_console


def enable_ssl(enable):
    config_dir = Path.cwd() / "core_system_config"

    if not config_dir.is_dir():
        rich_console.print("core_system_config directory does not exist")
        raise typer.Exit()

    # Update property files
    for property_path in config_dir.iterdir():
        with open(property_path, "r") as property_file:
            lines = property_file.readlines()
            update_line = f"server.ssl.enabled={str(enable).lower()}\n"
            updated_lines = [
                line if not line.startswith("server.ssl.enabled") else update_line
                for line in lines
            ]
        with open(property_path, "w") as property_file:
            property_file.writelines(updated_lines)

    with open(Path.cwd() / "cloud_config.yaml") as config_file:
        cloud_config = yaml.load(config_file, Loader=yamlloader.ordereddict.CSafeLoader)
        cloud_config["cloud"]["ssl_enabled"] = enable
    with open(Path.cwd() / "cloud_config.yaml", "w") as config_file:
        yaml.dump(cloud_config, config_file, Dumper=yamlloader.ordereddict.CSafeDumper)



from pathlib import Path
from typing import Union, List, Dict, Optional

import requests


def get_ssl_files(cloud_directory: Path):
    if (cloud_directory / "cloud_config.yaml").exists():
        cert_subpath = "certs/crypto/sysop.crt"
        key_subpath = "certs/crypto/sysop.key"
        ca_subpath = "certs/crypto/sysop.ca"
    else:
        cert_subpath = "sysop.crt"
        key_subpath = "sysop.key"
        ca_subpath = "sysop.ca"
    return (
        cloud_directory / subpath for subpath in (cert_subpath, key_subpath, ca_subpath)
    )


def get_service(
    url: str,
    cloud_directory: Path,
):
    *certkey, ca_path = get_ssl_files(cloud_directory)
    resp = requests.get(url, cert=certkey, verify=ca_path)
    return resp


def post_service(
    url: str, cloud_directory: Path, json: Union[Dict, List] = None, text: str = ""
):
    *certkey, ca_path = get_ssl_files(cloud_directory)
    if json:
        resp = requests.post(url, json=json, cert=certkey, verify=ca_path)
    elif text:
        resp = requests.post(url, data=text, cert=certkey, verify=ca_path)
    else:
        resp = requests.post(url, cert=certkey, verify=ca_path)
    return resp


def delete_service(
    url: str,
    cloud_directory: Path,
    params: Optional[Dict[str, str]] = None,
):
    *certkey, ca_path = get_ssl_files(cloud_directory)

    resp = requests.delete(url, params=params, cert=certkey, verify=ca_path)
    return resp



import json
from enum import Enum
from typing import Optional, Tuple, Dict

import typer
from rich import box
from rich.table import Table, Column

from pyrrowhead import rich_console
from pyrrowhead.management.serviceregistry import grouped_services
from pyrrowhead.management.utils import get_service, post_service, delete_service
from pyrrowhead.management.authorization import add_authorization_rule
from pyrrowhead.utils import (
    get_core_system_address_and_port,
    get_active_cloud_directory,
)


class SortbyChoices(str, Enum):
    RULE_ID = "id"
    PRIORITY = "priority"
    CONSUMER = "consumer"
    PROVIDER = "provider"
    SERVICE = "service"


def get_ids_from_service_definition(
    service_definition: str,
    interface_name: str,
    provider_name: str,
    address: str,
    port: int,
) -> Tuple[int, int, int]:
    """
    Args:
        service_definition:
        interface_name:
        provider_name:
        address:
        port:

    Returns:
        Tuple of service definition id, interface id, and provider id
    """
    response_data, status = grouped_services()

    services_by_definition = response_data.get("servicesGroupedByServiceDefinition")
    for service_definition_entry in services_by_definition:
        if service_definition_entry["serviceDefinition"] == service_definition:
            for service in service_definition_entry["providerServices"]:
                if (
                    service["provider"]["systemName"] == provider_name
                    and service["provider"]["address"] == address
                    and service["provider"]["port"] == port
                ):
                    correct_service = service
                    break
            else:
                rich_console.print(
                    f"Could not find provider {provider_name}@{address}:{port} for "
                    f"any service with service definition {service_definition}."
                )
                raise typer.Exit()
            break
    else:
        rich_console.print(
            f"Could not find any services with service definition {service_definition}."
        )
        raise typer.Exit()
    service_definition_id = correct_service["serviceDefinition"]["id"]
    provider_id = correct_service["provider"]["id"]
    interfaces = correct_service["interfaces"]
    for interface in interfaces:
        if interface["interfaceName"] == interface_name:
            interface_id = interface["id"]
            break
    else:
        rich_console.print(
            f"Could not find interface {interface_name} for service "
            f"{service_definition} in provider {provider_name}@{address}:{port}, "
            f"available interfaces are "
            f'{", ".join(interface["interfaceName"] for interface in interfaces)}'
        )
        raise typer.Exit()
    return service_definition_id, interface_id, provider_id


def table_sort(rule, choice):
    if choice == SortbyChoices.RULE_ID:
        return rule["id"]
    elif choice == SortbyChoices.PRIORITY:
        return rule["priority"]
    elif choice == SortbyChoices.CONSUMER:
        return rule["consumerSystem"]["id"]
    elif choice == SortbyChoices.PROVIDER:
        return rule["providerSystem"]["id"]
    elif choice == SortbyChoices.SERVICE:
        return rule["serviceDefinition"]["id"]
    else:
        raise RuntimeError(f"Invalid sorting choice: {choice}")


def table_condition(
    orch_rule: Dict,
    service_definition: Optional[str],
    consumer_id: Optional[int],
    consumer_name: Optional[str],
    provider_id: Optional[int],
    provider_name: Optional[str],
):
    if service_definition is not None:
        return orch_rule["serviceDefinition"]["serviceDefinition"] == service_definition
    elif consumer_id is not None:
        breakpoint()
        return orch_rule["consumerSystem"]["id"] == consumer_id
    elif consumer_name is not None:
        return orch_rule["consumerSystem"]["systemName"] == consumer_name
    elif provider_id is not None:
        return orch_rule["providerSystem"]["id"] == provider_id
    elif provider_name is not None:
        return orch_rule["providerSystem"]["systemName"] == provider_name
    else:
        return True


def create_orchestration_table(
    response_data: Dict,
    service_definition: Optional[str],
    consumer_id: Optional[int],
    consumer_name: Optional[str],
    provider_id: Optional[int],
    provider_name: Optional[str],
    sort_by: str,
):
    table = Table(
        Column(header="id", style="red"),
        Column(header="Consumer (id)", style="bright_blue"),
        Column(style="bright_blue"),
        Column(header="Provider (id)", style="blue"),
        Column(style="blue"),
        Column(header="Service Definition (id)", style="green"),
        Column(style="green"),
        Column(header="Interface (id)", style="bright_yellow"),
        Column(style="bright_yellow"),
        Column(header="priority", style="bright_white"),
        box=box.HORIZONTALS,
    )

    for orch_rule in sorted(
        response_data["data"], key=lambda x: table_sort(x, sort_by)
    ):
        if not table_condition(
            orch_rule,
            service_definition,
            consumer_id,
            consumer_name,
            provider_id,
            provider_name,
        ):
            continue
        table.add_row(
            str(orch_rule["id"]),
            f'{orch_rule["consumerSystem"]["systemName"]}',
            f'({orch_rule["consumerSystem"]["id"]})',
            f'{orch_rule["providerSystem"]["systemName"]}',
            f'({orch_rule["providerSystem"]["id"]})',
            f'{orch_rule["serviceDefinition"]["serviceDefinition"]}',
            f'({orch_rule["serviceDefinition"]["id"]})',
            f'{orch_rule["serviceInterface"]["interfaceName"]}',
            f'({orch_rule["serviceInterface"]["id"]})',
            str(orch_rule["priority"]),
        )

    return table


def list_orchestration_rules():
    active_cloud_directory = get_active_cloud_directory()
    address, port, secure, scheme = get_core_system_address_and_port(
        "orchestrator",
        active_cloud_directory,
    )
    scheme = "https" if secure else "http"
    response = get_service(
        f"{scheme}://{address}:{port}/orchestrator/mgmt/store",
        active_cloud_directory,
    )
    return response.json(), response.status_code


def add_orchestration_rule(
    service_definition: str,
    service_interface: str,
    provider_system: Tuple[str, str, int],
    consumer_id: Optional[int] = None,
    priority: int = 1,
    metadata: Optional[int] = None,
    add_auth_rule: Optional[bool] = None,
):
    active_cloud_directory = get_active_cloud_directory()
    address, port, secure, scheme = get_core_system_address_and_port(
        "orchestrator",
        active_cloud_directory,
    )
    # cloud_name = active_cloud_directory.name
    # org_name = active_cloud_directory.parents[0].name

    orchestration_input = [
        {
            "serviceDefinitionName": service_definition,
            "serviceInterfaceName": service_interface,
            "consumerSystemId": consumer_id,
            "providerSystem": dict(
                zip(("systemName", "address", "port"), provider_system)
            ),
            # "cloud": {"operator": org_name, "name": OPT_CLOUD_NAME, "secure": secure},
            "priority": priority,
            "attribute": metadata,
        }
    ]

    import pprint

    pprint.pprint(orchestration_input)
    response = post_service(
        f"{scheme}://{address}:{port}/orchestrator/mgmt/store",
        active_cloud_directory,
        json=orchestration_input,
    )

    if add_auth_rule:
        (
            service_definition_id,
            interface_id,
            provider_id,
        ) = get_ids_from_service_definition(
            service_definition, service_interface, *provider_system
        )
        add_authorization_rule(
            consumer_id,  # type: ignore
            provider_id,
            interface_id,
            service_definition_id
            # TODO: mypy complains here, remove ignore and fix later
        )

    try:
        response_data = response.json()
    except json.JSONDecodeError:
        response_data = {"errorMessage": "Error decoding json."}

    return response_data, response.status_code


def remove_orchestration_rule(orchestration_id):
    active_cloud_directory = get_active_cloud_directory()
    address, port, secure, scheme = get_core_system_address_and_port(
        "orchestrator",
        active_cloud_directory,
    )
    response = delete_service(
        f"{scheme}://{address}:{port}/orchestrator/mgmt/store/{orchestration_id}",
        active_cloud_directory,
    )

    try:
        response_data = response.json()
    except json.JSONDecodeError:
        response_data = {
            "errorMessage": "Could not decode orchestration "
            "response with status code {response.status_code"
        }
    return response_data, response.status_code



import shutil
import subprocess

import typer
import yaml
import yamlloader  # type: ignore
from rich.text import Text

from pyrrowhead.cloud.file_generators import generate_all_files
from pyrrowhead.cloud.initialize_cloud import initialize_cloud
from pyrrowhead import rich_console
from pyrrowhead.utils import get_config, set_config, PyrrowheadError


def install_cloud(config_file_path, installation_target):
    if config_file_path.suffix not in {".yaml", ".yml"}:
        raise PyrrowheadError("Configuration file must end with .yaml or .yml")
    elif not config_file_path.is_file():
        raise PyrrowheadError("Configuration file does not exist")

    with open(config_file_path, "r") as config_file:
        try:
            cloud_config = yaml.load(
                config_file, Loader=yamlloader.ordereddict.CSafeLoader
            )["cloud"]
        except (TypeError, KeyError):
            raise PyrrowheadError("Malformed configuration file")

    with rich_console.status(Text("Installing Arrowhead local cloud...")):
        generate_all_files(cloud_config, config_file_path, installation_target)
        initialize_cloud(
            installation_target,
            cloud_config["cloud_name"],
            cloud_config["organization_name"],
        )
    rich_console.print("Finished installing the [blue]Arrowhead[/blue] local cloud!")


def uninstall_cloud(
    installation_target, complete=False, keep_root=False, keep_sysop=False
):
    if not (installation_target / "cloud_config.yaml").exists():
        PyrrowheadError(
            f"{installation_target} does not contain an Arrowhead local cloud."
        )

    with open(installation_target / "cloud_config.yaml") as config_file:
        cloud_config = yaml.load(
            config_file, Loader=yamlloader.ordereddict.CSafeLoader
        )["cloud"]

    cloud_name = cloud_config["cloud_name"]
    org_name = cloud_config["organization_name"]

    if complete:
        # shutil.rmtree(installation_target)
        config = get_config()

        del config["local-clouds"][f"{cloud_name}.{org_name}"]
        set_config(config)
    else:
        if not keep_sysop:
            shutil.rmtree(installation_target / "certs")
        else:
            # TODO: Code that deletes everything except the sysop.* files
            pass
        shutil.rmtree(installation_target / "core_system_config")
        shutil.rmtree(installation_target / "sql")
        (installation_target / "docker-compose.yml").unlink()
        (installation_target / "initSQL.sh").unlink()
    subprocess.run(["docker", "volume", "rm", f"mysql.{cloud_name}.{org_name}"])
    rich_console.print("Uninstallation complete")
    raise typer.Exit()



from pathlib import Path
from collections import OrderedDict
from enum import Enum
import ipaddress

import yaml
import yamlloader  # type: ignore

from pyrrowhead.cloud.installation import install_cloud
from pyrrowhead.utils import get_config, set_config, validate_san
from pyrrowhead.types_ import ConfigDict


class CloudConfiguration(str, Enum):
    INTERCLOUD = "intercloud"
    EVENTHANDLER = "eventhandler"
    ONBOARDING = "onboarding"


def create_cloud_config(
    target_directory: Path,
    cloud_identifier,
    cloud_name,
    organization_name,
    ssl_enabled,
    ip_subnet,
    core_san,
    do_install,
    include,
):
    network = ipaddress.ip_network(ip_subnet)
    mandatory_core_systems = OrderedDict(
        {
            "service_registry": {
                "system_name": "service_registry",
                "address": str(network[3]),
                "domain": "serviceregistry",
                "port": 8443,
            },
            "orchestrator": {
                "system_name": "orchestrator",
                "address": str(network[4]),
                "domain": "orchestrator",
                "port": 8441,
            },
            "authorization": {
                "system_name": "authorization",
                "address": str(network[5]),
                "domain": "authorization",
                "port": 8445,
            },
        }
    )
    inter_cloud_core = OrderedDict(
        {
            "gateway": {
                "system_name": "gateway",
                "domain": "gateway",
                "port": 8453,
            },
            "gatekeeper": {
                "system_name": "gatekeeper",
                "domain": "gatekeeper",
                "port": 8449,
            },
        }
    )
    event_handling_core = {
        "event_handler": {
            "system_name": "event_handler",
            "domain": "eventhandler",
            "port": 8455,
        }
    }
    onboarding_core = OrderedDict(
        {
            "system_registry": {
                "system_name": "system_registry",
                "domain": "systemregistry",
                "port": 8437,
            },
            "device_registry": {
                "system_name": "device_registry",
                "domain": "deviceregistry",
                "port": 8439,
            },
            "certificate_authority": {
                "system_name": "certificate_authority",
                "domain": "certificate-authority",
                "port": 8448,
            },
            "onboarding_controller": {
                "system_name": "onboarding_controller",
                "domain": "onboarding-controller",
                "port": 8435,
            },
        }
    )

    cloud_core_services = mandatory_core_systems
    ip_start = len(mandatory_core_systems) + 3
    if CloudConfiguration.EVENTHANDLER in include:
        cloud_core_services.update(insert_ips(event_handling_core, network, ip_start))
        ip_start += len(event_handling_core)
    if CloudConfiguration.INTERCLOUD in include:
        cloud_core_services.update(insert_ips(inter_cloud_core, network, ip_start))
        ip_start += len(inter_cloud_core)
    if CloudConfiguration.ONBOARDING in include:
        cloud_core_services.update(insert_ips(onboarding_core, network, ip_start))
        ip_start += len(onboarding_core)

    for name in core_san:
        validate_san(name)

    cloud_config: ConfigDict = {
        "cloud": OrderedDict(  # type: ignore
            {
                "cloud_name": cloud_name,
                "organization_name": organization_name,
                "ssl_enabled": ssl_enabled,
                "subnet": str(network),
                "core_san": core_san,
                "client_systems": {},
                "core_systems": cloud_core_services,
            }
        )
    }

    target_directory = target_directory / f"{organization_name}/{cloud_name}"

    if not target_directory.is_absolute():
        target_directory = target_directory.expanduser()

    if not target_directory.exists():
        Path.mkdir(target_directory, parents=True)

    with open(
        target_directory / "cloud_config.yaml",
        "w",
    ) as yaml_file:
        yaml.dump(cloud_config, yaml_file, Dumper=yamlloader.ordereddict.CSafeDumper)

    config = get_config()
    config["local-clouds"][cloud_identifier] = str(target_directory)

    set_config(config)

    if do_install:
        install_cloud(target_directory / "cloud_config.yaml", target_directory)


def insert_ips(system_dict, network, start):
    return {
        sys_name: {**sys_data, **{"address": str(network[i])}}
        for i, (sys_name, sys_data) in enumerate(system_dict.items(), start=start)
    }



"""
Generates and copies files to the target directory.
"""
from pathlib import Path
import shutil
from functools import partial
from collections import OrderedDict
from importlib.resources import path
import ipaddress

from jinja2 import Environment, PackageLoader, select_autoescape
import yaml
import yamlloader  # type: ignore
from rich.text import Text

from pyrrowhead.database_config.passwords import db_passwords
from pyrrowhead import database_config
from pyrrowhead import rich_console
from pyrrowhead.types_ import CloudDict

yaml_safedump = partial(yaml.dump, Dumper=yamlloader.ordereddict.CSafeDumper)


def generate_config_files(cloud_config: CloudDict, target_path):
    """
    Creates the property files for all core services in yaml_path
    Args:
        yaml_path: Path to cloud config
        target_path: Path to cloud directory
    """
    env = Environment(
        loader=PackageLoader("pyrrowhead"), autoescape=select_autoescape()
    )

    core_systems = cloud_config["core_systems"]

    sr_address = core_systems["service_registry"]["address"]
    sr_port = core_systems["service_registry"]["port"]

    for system, config in core_systems.items():
        system_cn = (
            f'{config["system_name"]}.{cloud_config["cloud_name"]}.'
            f'{cloud_config["organization_name"]}.arrowhead.eu'
        )
        template = env.get_template(f"core_system_config/{system}.properties")

        system_config_file = template.render(
            **config,
            system_cn=system_cn,
            cloud_name=cloud_config["cloud_name"],
            organization_name=cloud_config["organization_name"],
            password=db_passwords[system],
            sr_address=sr_address,
            sr_port=sr_port,
            ssl_enabled=cloud_config["ssl_enabled"],
        )

        core_system_config_path = Path(target_path) / "core_system_config"
        core_system_config_path.mkdir(parents=True, exist_ok=True)
        with open(
            core_system_config_path / f"{system}.properties", "w+"
        ) as target_file:
            target_file.write(system_config_file)


def generate_docker_compose_file(cloud_config: CloudDict, target_path):
    cloud_identifier = (
        f'{cloud_config["cloud_name"]}.{cloud_config["organization_name"]}'
    )
    docker_compose_content = OrderedDict(
        {
            "version": "3",
            "services": OrderedDict(
                {
                    f"mysql.{cloud_identifier}": {
                        "container_name": f"mysql.{cloud_identifier}",
                        "image": "mysql:5.7",
                        "environment": ["MYSQL_ROOT_PASSWORD=123456"],
                        "volumes": [
                            f"mysql.{cloud_identifier}:/var/lib/mysql",
                            "./sql:/docker-entrypoint-initdb.d/",
                        ],
                        "networks": {
                            cloud_identifier: {
                                "ipv4_address": str(
                                    ipaddress.ip_network(cloud_config["subnet"])[2]
                                )
                            }
                        },
                        "ports": ["3306:3306"],
                    },
                }
            ),
            "volumes": {f"mysql.{cloud_identifier}": {"external": True}},
            "networks": {
                f"{cloud_identifier}": {
                    "ipam": {"config": [{"subnet": cloud_config["subnet"]}]}
                }
            },
        }
    )

    for core_system, config in cloud_config["core_systems"].items():
        core_name = config["domain"]
        # cloud_name = cloud_config["cloud_name"]
        docker_compose_content["services"][core_name] = {  # type: ignore
            "container_name": f"{core_name}.{cloud_identifier}",
            "image": f"svetlint/{core_name}:4.3.0",
            "depends_on": [f"mysql.{cloud_identifier}"],
            "volumes": [
                f"./core_system_config/{core_system}.properties:/"
                f"{core_name}/application.properties",
                f"./certs/crypto/{core_system}.p12:/{core_name}/{core_system}.p12",
                f"./certs/crypto/truststore.p12:/{core_name}/truststore.p12",
            ],
            "networks": {cloud_identifier: {"ipv4_address": config["address"]}},
            "ports": [f'{config["port"]}:{config["port"]}'],
        }

    with open(Path(target_path) / "docker-compose.yml", "w+") as target_file:
        yaml_safedump(docker_compose_content, target_file)


def generate_all_files(cloud_config, yaml_path, target_path):
    generate_config_files(cloud_config, target_path)
    rich_console.print(Text("Generated core system configuration files."))
    generate_docker_compose_file(cloud_config, target_path)
    rich_console.print(Text("Generated docker-compose.yml."))
    # Copy files that need not be generated
    with path(database_config, "initSQL.sh") as init_sql_path:
        shutil.copy(init_sql_path, target_path)
    # with path(certificate_generation, "lib_certs.sh") as lib_cert_path:
    #    shutil.copy(lib_cert_path, target_path / "certgen")
    # with path(certificate_generation, "rm_certs.sh") as rm_certs_path:
    #    shutil.copy(rm_certs_path, target_path / "certgen")
    # Copy the config file
    try:
        shutil.copy(yaml_path.absolute(), target_path)
    except shutil.SameFileError:
        pass
    rich_console.print(Text("Copied files."))



from rich import box
from rich.table import Table, Column

from pyrrowhead.management.utils import get_service, post_service
from pyrrowhead.utils import (
    get_core_system_address_and_port,
    get_active_cloud_directory,
)


def list_authorization_rules():
    active_cloud_directory = get_active_cloud_directory()
    address, port, secure, scheme = get_core_system_address_and_port(
        "authorization",
        active_cloud_directory,
    )
    response = get_service(
        f"{scheme}://{address}:{port}/authorization/mgmt/intracloud",
        active_cloud_directory,
    )
    response_data = response.json()
    status = response.status_code
    return response_data, status


def add_authorization_rule(
    consumer_id: int, provider_id: int, interface_id: int, service_definition_id: int
):
    active_cloud_directory = get_active_cloud_directory()
    address, port, secure, scheme = get_core_system_address_and_port(
        "authorization",
        active_cloud_directory,
    )

    rule_message = {
        "consumerId": consumer_id,
        "providerIds": [provider_id],
        "interfaceIds": [interface_id],
        "serviceDefinitionIds": [service_definition_id],
    }

    # TODO: use the response data
    response_data = post_service(  # noqa
        f"{scheme}://{address}:{port}/authorization/mgmt/intracloud/",
        active_cloud_directory,
        json=rule_message,
    )


def remove_authorization_rule():
    raise NotImplementedError


def create_authorization_table(response_data):
    auth_table = Table(
        Column(header="id", style="red"),
        Column(header="Consumer (id)", style="bright_blue"),
        Column(style="bright_blue"),
        Column(header="Provider (id)", style="blue"),
        Column(style="blue"),
        Column(header="Service definition (id)", style="green"),
        Column(style="green"),
        Column(header="Interface (id)", style="bright_yellow"),
        Column(style="bright_yellow"),
        title="Authorization rule",
        box=box.HORIZONTALS,
    )

    for auth_rule in response_data["data"]:
        row_data = [
            str(auth_rule["id"]),
            f'{auth_rule["consumerSystem"]["systemName"]}',
            f'(id: {auth_rule["consumerSystem"]["id"]})',
            f'{auth_rule["providerSystem"]["systemName"]}',
            f'(id: {auth_rule["providerSystem"]["id"]})',
            f'{auth_rule["serviceDefinition"]["serviceDefinition"]}',
            f'(id: {auth_rule["serviceDefinition"]["id"]})',
            f'{auth_rule["interfaces"][0]["interfaceName"]}',
            f'(id: {auth_rule["interfaces"][0]["id"]})',
        ]

        auth_table.add_row(*row_data)

    return auth_table



import subprocess

from rich.text import Text

from pyrrowhead import rich_console
from pyrrowhead.new_certificate_generation.generate_certificates import (
    setup_certificates,
)


def check_certs_exist(cloud_directory, cloud_name):
    cert_directory = cloud_directory / f"cloud-{cloud_name}/crypto"
    return cert_directory.is_dir() and any(cert_directory.iterdir())


def check_sql_initialized(cloud_directory):
    return (cloud_directory / "sql/create_empty_arrowhead_db.sql").is_file()


def check_mysql_volume_exists(cloud_name, org_name):
    ps_output = subprocess.run(
        ["docker", "volume", "ls"],
        capture_output=True,
    ).stdout.decode()
    # If mysql volume doesn't exists in stdout find returns -1
    return ps_output.find(f"mysql.{cloud_name}.{org_name}") != -1


def initialize_cloud(cloud_directory, cloud_name, organization_name):
    setup_certificates(cloud_directory / "cloud_config.yaml", "123456")
    rich_console.print(Text("Created certificates."))
    if not check_sql_initialized(cloud_directory):
        subprocess.run(["./initSQL.sh"], cwd=cloud_directory, capture_output=True)
        rich_console.print(Text("Initialized SQL tables."))
    if not check_mysql_volume_exists(cloud_name, organization_name):
        subprocess.run(
            f"docker volume create --name mysql.{cloud_name}."
            f"{organization_name}".split(),
            capture_output=True,
        )
        rich_console.print(Text("Created docker volume."))



import subprocess
from pathlib import Path

from rich.text import Text

from pyrrowhead import rich_console
from pyrrowhead.utils import switch_directory


def stop_local_cloud(cloud_directory: Path):
    with switch_directory(cloud_directory):
        with rich_console.status(Text("Stopping local cloud...")):
            subprocess.run(["docker-compose", "down"], capture_output=True)
        rich_console.print(Text("Local cloud Stopped."))



import os
from pathlib import Path
from contextlib import contextmanager
from typing import Tuple
import configparser
from ipaddress import ip_address

import typer
import yaml


class PyrrowheadError(Exception):
    pass


def get_config() -> configparser.ConfigParser:
    from pyrrowhead.constants import CONFIG_FILE

    config = configparser.ConfigParser()
    with open(get_pyrrowhead_path().joinpath(CONFIG_FILE), "r") as config_file:
        config.read_file(config_file)

    return config


def set_config(config: configparser.ConfigParser):
    from pyrrowhead.constants import CONFIG_FILE

    with open(get_pyrrowhead_path().joinpath(CONFIG_FILE), "w") as config_file:
        config.write(config_file)


@contextmanager
def switch_directory(path: Path):
    origin = Path.cwd()
    try:
        os.chdir(path)
        yield
    finally:
        os.chdir(origin)


def set_active_cloud(cloud_identifier):
    config = get_config()

    config["pyrrowhead"]["active-cloud"] = cloud_identifier

    set_config(config)


def get_active_cloud_directory() -> Path:
    config = get_config()

    active_cloud_identifier = config["pyrrowhead"]["active-cloud"]

    active_cloud_directory = config["local-clouds"][active_cloud_identifier]

    return Path(active_cloud_directory)


def get_core_system_address_and_port(
    core_system: str, cloud_directory: Path
) -> Tuple[str, int, bool, str]:
    from pyrrowhead.constants import CLOUD_CONFIG_FILE_NAME

    with open(cloud_directory / CLOUD_CONFIG_FILE_NAME, "r") as cloud_config_file:
        cloud_config = yaml.safe_load(cloud_config_file)
    address = cloud_config["cloud"]["core_systems"][core_system]["address"]
    port = cloud_config["cloud"]["core_systems"][core_system]["port"]
    secure = cloud_config["cloud"]["ssl_enabled"]
    scheme = "https" if secure else "http"

    return address, port, secure, scheme


def get_local_cloud_directory() -> Path:
    from pyrrowhead.constants import LOCAL_CLOUDS_SUBDIR

    return get_pyrrowhead_path().joinpath(LOCAL_CLOUDS_SUBDIR)


def get_pyrrowhead_path() -> Path:
    from pyrrowhead.constants import APP_NAME

    return Path(typer.get_app_dir(APP_NAME, force_posix=True))


def validate_san(san_candidate: str):
    if not (san_candidate.startswith("dns:") or san_candidate.startswith("ip:")):
        raise PyrrowheadError(
            "Subject Alternative Name must start with either 'ip:' or 'dns:'"
        )
    elif san_candidate.startswith("ip:"):
        if not check_valid_ip(san_candidate[3:]):
            raise PyrrowheadError(f"Malformed san ip: '{san_candidate}'")
    elif san_candidate.startswith("dns:"):
        if not check_valid_dns(san_candidate[4:]):
            raise PyrrowheadError(f"Malformed san dns: '{san_candidate}'")


def check_valid_ip(ip_candidate: str):
    try:
        ip_address(ip_candidate)
        return True
    except ValueError:
        return False


def check_valid_dns(identifier: str):
    import re

    identifier_re = re.compile(
        r"^(([a-zA-Z0-9]|[a-zA-Z0-9][a-zA-Z0-9\-]*[a-zA-Z0-9])\.)"
        r"*([A-Za-z0-9]|[A-Za-z0-9][A-Za-z0-9\-]*[A-Za-z0-9])$"
    )

    return identifier_re.search(identifier) is not None



from typing import Dict, TypedDict, List


class SystemDict(TypedDict):
    system_name: str
    address: str
    port: int


class ClientSystemDict(SystemDict):
    sans: List[str]


class CoreSystemDict(SystemDict):
    domain: str


class CloudDict(TypedDict):
    cloud_name: str
    organization_name: str
    ssl_enabled: bool
    subnet: str
    core_san: List[str]
    core_systems: Dict[str, CoreSystemDict]
    client_systems: Dict[str, ClientSystemDict]


class ConfigDict(TypedDict):
    cloud: CloudDict



from typing import Optional, Tuple, Dict

from rich import box
from rich.console import RenderGroup
from rich.panel import Panel
from rich.table import Table, Column

from pyrrowhead import rich_console
from pyrrowhead.management.common import AccessPolicy
from pyrrowhead.management.utils import (
    get_service,
    post_service,
    delete_service as del_service,
)
from pyrrowhead.utils import (
    get_core_system_address_and_port,
    get_active_cloud_directory,
)


def inspect_service(service_id: int):
    active_cloud_directory = get_active_cloud_directory()
    address, port, secure, scheme = get_core_system_address_and_port(
        "service_registry",
        active_cloud_directory,
    )

    response = get_service(
        f"{scheme}://{address}:{port}/serviceregistry/mgmt/{service_id}",
        active_cloud_directory,
    )

    return response.json(), response.status_code


def add_service(
    service_definition: str,
    uri: str,
    interface: str,
    access_policy: AccessPolicy,
    system: Tuple[str, str, int],
):
    active_cloud_directory = get_active_cloud_directory()
    sr_address, sr_port, secure, scheme = get_core_system_address_and_port(
        "service_registry",
        active_cloud_directory,
    )

    system_name, address, port = system

    registry_request = {
        "serviceDefinition": service_definition,
        "serviceUri": uri,
        "interfaces": [interface],
        "secure": access_policy,
        "providerSystem": {
            "systemName": system_name,
            "address": address,
            "port": port,
        },
    }

    response = post_service(
        f"{scheme}://{sr_address}:{sr_port}/serviceregistry/mgmt/",
        active_cloud_directory,
        json=registry_request,
    )

    return response.json(), response.status_code


def list_filter(
    orch_rule: Dict,
    service_definition: Optional[str],
    system_name: Optional[str],
    system_id: Optional[int],
):
    if service_definition is not None:
        return orch_rule["serviceDefinition"]["serviceDefinition"] == service_definition
    elif system_name is not None:
        return orch_rule["provider"]["systemName"] == system_name
    elif system_id is not None:
        return orch_rule["provider"]["id"] == system_id

    return True


def list_services(
    service_definition: Optional[str],
    system_name: Optional[str],
    system_id: Optional[int],
) -> Dict:
    active_cloud_directory = get_active_cloud_directory()
    address, port, secure, scheme = get_core_system_address_and_port(
        "service_registry",
        active_cloud_directory,
    )

    endpoint = f"{scheme}://{address}:{port}/serviceregistry/mgmt/"

    response_data = get_service(
        endpoint,
        active_cloud_directory,
    ).json()

    if any((system_name, system_id)):
        response_data = {
            "data": [
                service
                for service in response_data["data"]
                if (
                    system_name == service["provider"]["systemName"]
                    or system_id == service["provider"]["id"]
                )
            ]
        }

    response_data["data"] = [
        orch_rule
        for orch_rule in response_data["data"]
        if list_filter(orch_rule, service_definition, system_name, system_id)
    ]

    return response_data


def delete_service(service_id: int):
    active_cloud_directory = get_active_cloud_directory()
    address, port, secure, scheme = get_core_system_address_and_port(
        "service_registry",
        active_cloud_directory,
    )

    response = del_service(
        f"https://{address}:{port}/serviceregistry/mgmt/{id}",
        active_cloud_directory,
    )

    return response.json(), response.status_code


def create_service_table(
    response_data: Dict, show_system, show_access_policy, show_service_uri
) -> Table:
    service_table = Table(
        Column(header="id", style="red"),
        Column(header="Service definition", style="bright_blue"),
        Column(header="Interface", style="green"),
        title="Registered Services",
        box=box.SIMPLE,
    )

    if show_service_uri:
        service_table.add_column(header="Service URI", style="bright_yellow")
    if show_access_policy:
        service_table.add_column(
            header="Access Policy",
            style="orange3",
        )
    if show_system:
        service_table.add_column(
            header="System",
            style="blue",
        )

    for service in response_data["data"]:
        row_data = [
            str(service["id"]),
            f'{service["serviceDefinition"]["serviceDefinition"]}  '
            f'(id: {service["serviceDefinition"]["id"]})',
            service["interfaces"][0]["interfaceName"],
        ]

        if show_service_uri:
            row_data.append(service["serviceUri"])
        if show_access_policy:
            row_data.append(service["secure"])
        if show_system:
            row_data.append(
                f'{service["provider"]["systemName"]}  '
                f'(id: {service["provider"]["id"]})'
            )
        service_table.add_row(*row_data)

    return service_table


def grouped_services():
    active_cloud_directory = get_active_cloud_directory()
    address, port, secure, scheme = get_core_system_address_and_port(
        "service_registry",
        active_cloud_directory,
    )
    response = get_service(
        f"{scheme}://{address}:{port}/serviceregistry/mgmt/grouped",
        active_cloud_directory,
    )
    return response.json(), response.status_code


def get_system_id_from_name(system_name: str, address: str = "", port: int = -1) -> int:
    active_cloud_directory = get_active_cloud_directory()
    sr_address, sr_port, secure, scheme = get_core_system_address_and_port(
        "service_registry",
        active_cloud_directory,
    )

    scheme = "https" if secure else "http"
    response_data = get_service(
        f"{scheme}://{sr_address}:{sr_port}/serviceregistry/mgmt/systems",
        active_cloud_directory,
    ).json()

    candidate_systems = [
        system
        for system in response_data["data"]
        if system["systemName"] == system_name
    ]

    if len(address) > 0 and port >= 0:
        candidate_systems = [
            system
            for system in candidate_systems
            if system["address"] == address and system["port"] == port
        ]

    if len(candidate_systems) == 0:
        return -1
    elif len(candidate_systems) > 1:
        return -2

    return candidate_systems[0]["id"]


def render_service(response_data):
    tab_break = "\n\t"
    provider = response_data["provider"]
    render_group = RenderGroup(
        (f'Service URI: {response_data["serviceUri"]}'),
        (
            f"Interfaces: [bright_yellow]\n\t"
            f'{tab_break.join(interface["interfaceName"] for interface in response_data["interfaces"])}'  # noqa
            f"[/bright_yellow]"
        ),
        (f'Access policy: [orange3]{response_data["secure"]}[/orange3]'),
        (f'Version: {response_data["version"]}'),
        (
            f"Provider:"
            f'{tab_break}Id: {provider["id"]}'
            f'{tab_break}System name: [blue]{provider["systemName"]}[/blue]'
            f'{tab_break}Address: {provider["address"]}'
            f'{tab_break}Port: {provider["port"]}'
        ),
        (
            f"End of validity: [red]"
            f'{response_data.get("endOfValidity", "[green]Always valid[/green]")}'
            f"[/red]"
        ),
    )
    if metadata := response_data.get("metadata"):
        render_group.renderables.append(
            f"Metadata: {tab_break} "
            f'{tab_break.join(f"{name}: {value}" for name, value in metadata.items())}'
        )
    rich_console.print(
        (
            f' Service "{response_data["serviceDefinition"]["serviceDefinition"]}" '
            f'(id: {response_data["id"]})'
        ),
        Panel(render_group, box=box.HORIZONTALS, expand=False),
    )



import json
from typing import Tuple, Optional

import typer
from rich.syntax import Syntax

from pyrrowhead.management import common, serviceregistry, orchestrator
from pyrrowhead import rich_console

orch_app = typer.Typer(name="orchestration")


@orch_app.command(name="add")
def add_orchestration_rule_cli(
    service_definition: str = common.ARG_SERVICE_DEFINITION,
    service_interface: str = common.ARG_SERVICE_INTERFACE,
    provider: Tuple[str, str, int] = typer.Option(
        ...,
        show_default=False,
        metavar="SYSTEM_NAME ADDRESS PORT",
        help="Provider system definition.",
    ),
    consumer_id: Optional[int] = typer.Option(
        None, metavar="CONSUMER_ID", help="Incompatible with " "CONSUMER option"
    ),  # Union[int, str, Tuple[str, str, int]] = typer.Option(...),
    consumer: Tuple[str, str, int] = typer.Option(
        (None, None, None),
        show_default=False,
        metavar="SYSTEM_NAME ADDRESS PORT",
        help="Consumer system definition, " "incompatible with CONSUMER_ID option.",
    ),
    priority: int = typer.Option(1),
    add_auth_rule: Optional[bool] = typer.Option(
        None,
        "--add-authentication",
        "-A",
        help="Add authentication rule in together " "with the authentication rule",
    ),
):
    if consumer_id is not None:
        pass
    elif not all(consumer):
        consumer_id = serviceregistry.get_system_id_from_name(*consumer)
    elif all(consumer):
        consumer_id = serviceregistry.get_system_id_from_name(*consumer)
    else:
        rich_console.print(
            "No consumer information given, you must provide "
            "Pyrrowhead with either the consumer id (--consumer-id), "
            "consumer name (--consumer-name) or "
            "full consumer information (--consumer-system)."
        )

    if consumer_id == -1:
        rich_console.print(f"No consumer systems found for consumer {consumer[0]}")
        raise typer.Exit()
    if consumer_id == -2:
        rich_console.print(
            f"Multiple candidate systems found for consumer {consumer[0]}, "
            f"please specify address and port"
        )
        raise typer.Exit()

    response_data, status = orchestrator.add_orchestration_rule(
        service_definition=service_definition,
        service_interface=service_interface,
        provider_system=provider,
        consumer_id=consumer_id,
        priority=priority,
        add_auth_rule=add_auth_rule,
    )

    if status >= 400:
        print(response_data["errorMessage"], status)


@orch_app.command(name="list")
def list_orchestration_cli(
    service_definition: Optional[str] = typer.Option(
        None, metavar="SERVICE_DEFINITION"
    ),
    provider_id: Optional[int] = typer.Option(None),
    provider_name: Optional[str] = typer.Option(None),
    consumer_id: Optional[int] = typer.Option(None),
    consumer_name: Optional[str] = typer.Option(None),
    sort_by: orchestrator.SortbyChoices = typer.Option("id"),
    raw_output: bool = common.OPT_RAW_OUTPUT,
    raw_indent: Optional[int] = common.OPT_RAW_INDENT,
):
    response_data, status = orchestrator.list_orchestration_rules()

    if raw_output:
        rich_console.print(Syntax(json.dumps(response_data, indent=raw_indent), "json"))
        raise typer.Exit()

    table = orchestrator.create_orchestration_table(
        response_data,
        service_definition,
        consumer_id,
        consumer_name,
        provider_id,
        provider_name,
        sort_by,
    )

    rich_console.print(table)


@orch_app.command(name="remove")
def remove_orchestration_cli(orchestration_id: int):
    response_data, status = orchestrator.remove_orchestration_rule(orchestration_id)



from pathlib import Path
from typing import List, Optional
import ipaddress

import yaml
import yamlloader  # type: ignore

from pyrrowhead.types_ import CloudDict, ClientSystemDict
from pyrrowhead.utils import (
    PyrrowheadError,
    check_valid_dns,
    validate_san,
    check_valid_ip,
)


def find_first_missing(ints: List[int], start: int, stop: int) -> int:
    for i in range(start, stop):
        if i not in set(ints):
            break

    return i


def add_client_system(
    config_file_path: Path,
    system_name: str,
    system_address: Optional[str],
    system_port: Optional[int],
    system_additional_addresses: Optional[List[str]],
):
    with open(config_file_path, "r") as config_file:
        cloud_config: CloudDict = yaml.load(
            config_file, Loader=yamlloader.ordereddict.CSafeLoader
        )["cloud"]

    if system_address is not None and check_valid_ip(system_address):
        addr = system_address
    elif system_address is not None:
        raise PyrrowheadError(
            f"System address '{system_address}' " f"is not a valid ip address."
        )
    else:
        addr = str(ipaddress.ip_network(cloud_config["subnet"])[1])

    if not check_valid_dns(system_name):
        raise PyrrowheadError(
            f"System name '{system_name}' " f"is not a valid dns string."
        )

    if system_additional_addresses is not None:
        for name in system_additional_addresses:
            validate_san(name)

    for sys in cloud_config.get("client_systems", {}).values():
        if (
            sys["system_name"] == system_name
            and sys["address"] == system_address
            and sys["port"] == system_port
        ):
            raise PyrrowheadError(
                f'Client system with name "{system_name}", '
                f"address {system_address}, and port {system_address} already exists"
            )

    taken_ports = [
        sys["port"]
        for sys in cloud_config.get("client_systems", {}).values()
        if sys["address"] == addr
    ]
    if system_port is None or system_port in taken_ports:
        port = find_first_missing(taken_ports, 5000, 8000)
    else:
        port = system_port

    id = (
        system_name
        + "-"
        + str(
            len(
                tuple(
                    sys
                    for sys in cloud_config["client_systems"].values()
                    if sys["system_name"] == system_name
                )
            )
        ).rjust(3, "0")
    )

    system_dict: ClientSystemDict = {
        "system_name": system_name,
        "address": addr,
        "port": port,
        "sans": [],
    }

    if system_additional_addresses is not None:
        system_dict["sans"] = system_additional_addresses

    cloud_config["client_systems"][id] = system_dict

    with open(config_file_path, "w") as config_file:
        yaml.dump(
            {"cloud": cloud_config},
            config_file,
            Dumper=yamlloader.ordereddict.CSafeDumper,
        )



import typer

from pyrrowhead.utils import get_local_cloud_directory

# String constants
ENV_PYRROWHEAD_DIRECTORY = "PYRROWHEAD_INSTALL_DIRECTORY"
ENV_PYRROWHEAD_ACTIVE_CLOUD = "PYRROWHEAD_ACTIVE_CLOUD"
APP_NAME = "pyrrowhead"
LOCAL_CLOUDS_SUBDIR = "local-clouds"
CLOUD_CONFIG_FILE_NAME = "cloud_config.yaml"
CONFIG_FILE = "config.cfg"
ORG_CERT_DIR = "org-certs"
ROOT_CERT_DIR = "root-certs"

# Typer constants
OPT_CLOUDS_DIRECTORY = typer.Option(
    None,
    "--dir",
    "-d",
    callback=get_local_cloud_directory,
    help="Directory of local cloud. Experimental feature. "
    "Should only be used when a local cloud is "
    "installed outside the default path.",
)
ARG_CLOUD_IDENTIFIER = typer.Argument(
    None,
    help="""
Cloud identifier string of format <CLOUD_NAME>.<ORG_NAME>.
Mutually exclusive with options -c and -o.
""",
)
ARG_ORG_NAME = typer.Argument(
    None,
    help="""
    Organization name.
    """,
)
OPT_CLOUD_NAME = typer.Option(
    None,
    "--cloud",
    "-c",
    help="CLOUD_NAME. Mandatory with option -o and "
    "mutually exclusive with argument CLOUD_IDENTIFIER",
)
OPT_ORG_NAME = typer.Option(
    None,
    "--org",
    "-o",
    help="ORG_NAME. Mandatory with option -c and "
    "mutually exclusive with argument CLOUD_IDENTIFIER",
)



from pathlib import Path
import socket
import ssl
import subprocess
import time

import yaml
import yamlloader  # type: ignore
from rich.text import Text
import typer

from pyrrowhead import rich_console
from pyrrowhead.cloud.stop import stop_local_cloud
from pyrrowhead.utils import switch_directory, PyrrowheadError


def check_server(address, port, secure, certfile, keyfile, cafile):
    if secure:
        context = ssl.create_default_context(cafile=cafile.absolute())
        context.verify_mode = ssl.CERT_REQUIRED
        context.check_hostname = False
        try:
            context.load_cert_chain(certfile, keyfile)
        except OSError:
            raise PyrrowheadError(
                "Could not load certificates in <pyrrowhead.cloud.start.check_server>."
            )

        try:
            with socket.create_connection((address, port)) as sock:
                with context.wrap_socket(sock, server_hostname=address):
                    return True
        except (
            ConnectionRefusedError,
            ConnectionResetError,
            OSError,
        ):
            return False
    else:
        try:
            with socket.create_connection((address, port)) as sock:
                return True
        except (ConnectionRefusedError, ConnectionResetError, OSError):
            return False


def is_port_in_use(port):
    import socket

    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
        return s.connect_ex(("localhost", port)) == 0


def check_returncode(output, status, cloud_directory: Path = Path.cwd()):
    if output.returncode != 0:
        rich_console.print(Text("Encountered an error during startup: "))
        rich_console.print(Text(output.stderr.decode()))
        status.stop()
        stop_local_cloud(cloud_directory)
        raise typer.Abort()


def start_local_cloud(cloud_directory: Path):
    with open(cloud_directory / "cloud_config.yaml") as config_file:
        cloud_config = yaml.load(config_file, Loader=yamlloader.ordereddict.Loader)
    cloud_name = cloud_config["cloud"]["cloud_name"]
    org_name = cloud_config["cloud"]["organization_name"]
    ssl_enabled = cloud_config["cloud"]["ssl_enabled"]

    sysop_certfile = (cloud_directory / "certs/crypto/sysop.crt").absolute()
    sysop_keyfile = (cloud_directory / "certs/crypto/sysop.key").absolute()
    sysop_cafile = (cloud_directory / "certs/crypto/sysop.ca").absolute()
    core_systems = cloud_config["cloud"]["core_systems"]

    with switch_directory(cloud_directory):
        with rich_console.status("MySQL instance starting...") as status:
            output = subprocess.run(
                ["docker-compose", "up", "-d", f"mysql.{cloud_name}.{org_name}"],
                capture_output=True,
            )
            check_returncode(output, status)
            rich_console.print(Text("MySQL instance started."))
            for core_system, core_system_config in core_systems.items():
                core_system_print_name = (
                    core_system_config["system_name"].replace("_", " ").capitalize()
                )
                status.update(Text(f"{core_system_print_name} starting..."))
                output = subprocess.run(
                    ["docker-compose", "up", "-d", core_system_config["domain"]],
                    capture_output=True,
                )
                check_returncode(output, status)
                while not check_server(
                    core_system_config["address"],
                    core_system_config["port"],
                    ssl_enabled,
                    sysop_certfile,
                    sysop_keyfile,
                    sysop_cafile,
                ):
                    time.sleep(1)
                rich_console.print(Text(f"{core_system_print_name} started."))
            rich_console.print("Local cloud is up and running!")



import typer

from pyrrowhead.management.cli import sr_app, orch_app, auth_app, sys_app
from pyrrowhead.cloud.cli import cloud_app

# from pyrrowhead.org.cli import org_app
from pyrrowhead._setup import _setup_pyrrowhead

app = typer.Typer(callback=_setup_pyrrowhead)
app.add_typer(sr_app)
app.add_typer(orch_app)
app.add_typer(auth_app)
app.add_typer(sys_app)
app.add_typer(cloud_app)
# The org command is work in progress
# app.add_typer(org_app)


if __name__ == "__main__":
    app()



from pathlib import Path
from typing import Tuple
import configparser

import typer

from pyrrowhead import utils, rich_console
from pyrrowhead.constants import APP_NAME, LOCAL_CLOUDS_SUBDIR, CONFIG_FILE


def _is_initialized(pyrrowhead_path: Path) -> Tuple[bool, bool, bool]:
    path_exists = pyrrowhead_path.is_dir()
    config_exists = pyrrowhead_path.joinpath(CONFIG_FILE).is_file()
    cloud_dir_exists = pyrrowhead_path.joinpath(LOCAL_CLOUDS_SUBDIR).is_dir()
    return path_exists, config_exists, cloud_dir_exists


def _setup_pyrrowhead():
    pyrrowhead_path = utils.get_pyrrowhead_path()
    config = configparser.ConfigParser()

    path_exists, config_exists, cloud_dir_exists = _is_initialized(pyrrowhead_path)

    if all((path_exists, config_exists, cloud_dir_exists)):
        return

    if not path_exists:
        rich_console.print("Initializing pyrrowhead directory.")
        pyrrowhead_path.mkdir()
    elif len(list(pyrrowhead_path.iterdir())) == 0:
        rich_console.print("Empty pyrrowhead directory found.")
    elif len(list(pyrrowhead_path.iterdir())) > 2 or not (
        config_exists and cloud_dir_exists
    ):
        rich_console.print(
            "Pyrrowhead directory contains unknown objects, initialization aborted."
        )
        raise typer.Exit(-1)

    if not config_exists and not cloud_dir_exists:
        rich_console.print("Initializing config file.")
        config[APP_NAME] = {}
        config[LOCAL_CLOUDS_SUBDIR] = {}
        utils.set_config(config)
        rich_console.print("Initializing empty local cloud directory.")
        pyrrowhead_path.joinpath(LOCAL_CLOUDS_SUBDIR).mkdir()
    elif config_exists and not cloud_dir_exists:
        rich_console.print("Initializing empty local cloud directory.")
        pyrrowhead_path.joinpath(LOCAL_CLOUDS_SUBDIR).mkdir()
    elif not config_exists and cloud_dir_exists:
        rich_console.print("Initializing config file.")
        for org_path in pyrrowhead_path.joinpath(LOCAL_CLOUDS_SUBDIR).iterdir():
            for cloud_path in org_path.iterdir():
                config["local-clouds"][
                    ".".join((str(cloud_path.name), str(org_path.name)))
                ] = str(cloud_path.absolute())

        utils.set_config(config)

    rich_console.print("Initialization completed.")



from pyrrowhead.management.serviceregistry import (
    inspect_service,
    add_service,
    list_services,
    delete_service,
    grouped_services,
    get_system_id_from_name,
)
from pyrrowhead.management.orchestrator import (
    add_orchestration_rule,
    list_orchestration_rules,
    remove_orchestration_rule,
)
from pyrrowhead.management.authorization import (
    add_authorization_rule,
    list_authorization_rules,
    remove_authorization_rule,
)
from pyrrowhead.management.systemregistry import (
    add_system,
    list_systems,
    remove_system,
)

__all__ = [
    # SERVICE REGISTRY
    "inspect_service",
    "add_service",
    "list_services",
    "delete_service",
    "grouped_services",
    "get_system_id_from_name",
    # ORCHESTRATOR
    "add_orchestration_rule",
    "list_orchestration_rules",
    "remove_orchestration_rule",
    # AUTHORIZATION
    "add_authorization_rule",
    "list_authorization_rules",
    "remove_authorization_rule",
    # SYSTEM REGISTRY
    "add_system",
    "list_systems",
    "remove_system",
]
